from terra_sdk.core.coins import Coins

from ._base import BaseAPI


class AsyncSupplyAPI(BaseAPI):
    async def total(self) -> Coins:
        res = await self._c._get("/supply/total")
        return Coins.from_data(res)


class SupplyAPI(BaseAPI):
    def total(self) -> Coins:
        """Fetches the current total supply of all tokens.

        Returns:
            Coins: total supply
        """
        res = self._c._get("/supply/total")
        return Coins.from_data(res)



from __future__ import annotations

import attr

from terra_sdk.core import Coins
from terra_sdk.util.base import BaseTerraData
from terra_sdk.util.json import JSONSerializable

__all__ = [
    "Authorization",
    "SendAuthorization",
    "GenericAuthorization",
    "AuthorizationGrant",
]


class Authorization(BaseTerraData):
    @staticmethod
    def from_data(data: dict) -> Authorization:
        from terra_sdk.util.parse_authorization import parse_authorization

        return parse_authorization(data)


@attr.s
class SendAuthorization(Authorization):
    type = "msgauth/SendAuthorization"

    spend_limit: Coins = attr.ib(converter=Coins)

    @classmethod
    def from_data(cls, data: dict) -> SendAuthorization:
        data = data["value"]
        return cls(spend_limit=Coins.from_data(data["spend_limit"]))


@attr.s
class GenericAuthorization(Authorization):
    type = "msgauth/GenericAuthorization"

    grant_msg_type: str = attr.ib()

    @classmethod
    def from_data(cls, data: dict) -> GenericAuthorization:
        data = data["value"]
        return cls(grant_msg_type=data["grant_msg_type"])


@attr.s
class AuthorizationGrant(JSONSerializable):

    authorization: Authorization = attr.ib()
    expiration: str = attr.ib()

    @classmethod
    def from_data(cls, data: dict) -> AuthorizationGrant:
        return cls(
            authorization=Authorization.from_data(data["authorization"]),
            expiration=data["expiration"],
        )



"""Some useful base classes to inherit from."""

from typing import Any, Callable, Dict, List

from .json import JSONSerializable, dict_to_data


class BaseTerraData(JSONSerializable):

    type: str

    def to_data(self) -> dict:
        return {"type": self.type, "value": dict_to_data(self.__dict__)}


def create_demux(inputs: List) -> Callable[[Dict[str, Any]], Any]:
    table = {i.type: i.from_data for i in inputs}

    def from_data(data: dict):
        return table[data["type"]](data)

    return from_data



from terra_sdk.core import Dec

from ._base import BaseAPI


class AsyncMintAPI(BaseAPI):
    async def inflation(self) -> Dec:
        return Dec(await self._c._get("/minting/inflation"))

    async def annual_provisions(self) -> Dec:
        return Dec(await self._c._get("/minting/annual-provisions"))

    async def parameters(self) -> dict:
        return await self._c._get("/minting/parameters")


class MintAPI(BaseAPI):
    def inflation(self) -> Dec:
        """Fetches the current inflation.

        Returns:
            Dec: inflation
        """
        return Dec(self._c._get("/minting/inflation"))

    def annual_provisions(self) -> Dec:
        """Fetches the annual provisions.

        Returns:
            Dec: annual provisions
        """
        return Dec(self._c._get("/minting/annual-provisions"))

    def parameters(self) -> dict:
        """Fetches the Mint module's parameters.

        Returns:
            dict: Mint module parameters
        """
        return self._c._get("/minting/parameters")



from typing import Optional

from ._base import BaseAPI


class AsyncTendermintAPI(BaseAPI):
    async def node_info(self) -> dict:
        return await self._c._get("/node_info", raw=True)

    async def syncing(self) -> bool:
        return (await self._c._get("/syncing", raw=True))["syncing"]

    async def validator_set(self, height: Optional[int] = None) -> dict:
        x = "latest" if height is None else height
        return await self._c._get(f"/validatorsets/{x}")

    async def block_info(self, height: Optional[int] = None) -> dict:
        x = "latest" if height is None else height
        return await self._c._get(f"/blocks/{x}", raw=True)


class TendermintAPI(BaseAPI):
    def node_info(self) -> dict:
        """Fetches the curent connected node's information.

        Returns:
            dict: node information
        """
        return self._c._get("/node_info", raw=True)

    def syncing(self) -> bool:
        """Fetches whether the curent connect node is syncing with the network.

        Returns:
            bool: syncing status
        """
        return (self._c._get("/syncing", raw=True))["syncing"]

    def validator_set(self, height: Optional[int] = None) -> dict:
        """Fetches the validator set for a height. If no height is given, defaults to latest.

        Args:
            height (Optional[int], optional): block height. Defaults to None.

        Returns:
            dict: validator set
        """
        x = "latest" if height is None else height
        return self._c._get(f"/validatorsets/{x}")

    def block_info(self, height: Optional[int] = None) -> dict:
        """Fetches the block information for a given height. If no height is given, defaults to latest block.

        Args:
            height (Optional[int], optional): block height. Defaults to None.

        Returns:
            dict: block info
        """
        x = "latest" if height is None else height
        return self._c._get(f"/blocks/{x}", raw=True)



from __future__ import annotations

import attr

from terra_sdk.core import AccAddress, Coins
from terra_sdk.core.msg import Msg

from .data import Content

__all__ = ["MsgSubmitProposal", "MsgDeposit", "MsgVote"]


@attr.s
class MsgSubmitProposal(Msg):

    type = "gov/MsgSubmitProposal"
    action = "submit_proposal"

    content: Content = attr.ib()
    initial_deposit: Coins = attr.ib(converter=Coins)
    proposer: AccAddress = attr.ib()

    @classmethod
    def from_data(cls, data: dict) -> MsgSubmitProposal:
        from terra_sdk.util.parse_content import parse_content

        data = data["value"]
        content = parse_content(data["content"])
        return cls(
            content=content,
            initial_deposit=Coins.from_data(data["initial_deposit"]),
            proposer=data["proposer"],
        )


@attr.s
class MsgDeposit(Msg):

    type = "gov/MsgDeposit"
    action = "deposit"

    proposal_id: int = attr.ib(converter=int)
    depositor: AccAddress = attr.ib()
    amount: Coins = attr.ib(converter=Coins)

    def to_data(self) -> dict:
        return {
            "type": self.type,
            "value": {
                "proposal_id": str(self.proposal_id),
                "depositor": self.depositor,
                "amount": self.amount.to_data(),
            },
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgDeposit:
        data = data["value"]
        return cls(
            proposal_id=data["proposal_id"],
            depositor=data["depositor"],
            amount=Coins.from_data(data["amount"]),
        )


@attr.s
class MsgVote(Msg):

    type = "gov/MsgVote"
    action = "vote"

    EMPTY = "Empty"
    YES = "Yes"
    ABSTAIN = "Abstain"
    NO = "No"
    NO_WITH_VETO = "NoWithVeto"

    proposal_id: int = attr.ib(converter=int)
    voter: AccAddress = attr.ib()
    option: str = attr.ib()

    @option.validator
    def _check_option(self, attribute, value):
        possible_options = [
            self.EMPTY,
            self.YES,
            self.ABSTAIN,
            self.NO,
            self.NO_WITH_VETO,
        ]
        if value not in possible_options:
            raise TypeError(
                f"incorrect value for option: {value}, must be one of: {possible_options}"
            )

    def to_data(self) -> dict:
        return {
            "type": self.type,
            "value": {
                "proposal_id": str(self.proposal_id),
                "voter": self.voter,
                "option": self.option,
            },
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgVote:
        data = data["value"]
        return cls(
            proposal_id=data["proposal_id"],
            voter=data["voter"],
            option=data["option"],
        )



__all__ = [
    "Coin",
    "Coins",
    "Dec",
    "Numeric",
    "PublicKey",
    "AccAddress",
    "AccPubKey",
    "ValAddress",
    "ValConsPubKey",
    "ValPubKey",
]

from .coin import Coin
from .coins import Coins
from .numeric import Dec, Numeric
from .public_key import PublicKey
from .strings import AccAddress, AccPubKey, ValAddress, ValConsPubKey, ValPubKey



from __future__ import annotations

from typing import NewType

from bech32 import bech32_decode, bech32_encode

__all__ = [
    "AccAddress",
    "ValAddress",
    "AccPubKey",
    "ValPubKey",
    "ValConsPubKey",
    "is_acc_address",
    "is_acc_pubkey",
    "is_val_address",
    "is_val_pubkey",
    "is_valcons_pubkey",
    "to_acc_address",
    "to_acc_pubkey",
    "to_val_address",
    "to_val_pubkey",
]


def check_prefix_and_length(prefix: str, data: str, length: int):
    vals = bech32_decode(data)
    return vals[0] == prefix and len(data) == length


AccAddress = NewType("AccAddress", str)
ValAddress = NewType("ValAddress", str)
AccPubKey = NewType("AccPubKey", str)
ValPubKey = NewType("ValPubKey", str)
ValConsPubKey = NewType("ValConsPubKey", str)


def is_acc_address(data: str) -> bool:
    return check_prefix_and_length("terra", data, 44)


def to_acc_address(data: ValAddress) -> AccAddress:
    vals = bech32_decode(data)
    if vals[1] is None:
        raise ValueError(f"invalid bech32: {data}")
    return AccAddress(bech32_encode("terra", vals[1]))


def is_val_address(data: str) -> bool:
    return check_prefix_and_length("terravaloper", data, 51)


def to_val_address(data: AccAddress) -> ValAddress:
    vals = bech32_decode(data)
    if vals[1] is None:
        raise ValueError(f"invalid bech32: {data}")
    return ValAddress(bech32_encode("terravaloper", vals[1]))


def is_acc_pubkey(data: str) -> bool:
    return check_prefix_and_length("terrapub", data, 76)


def to_acc_pubkey(data: ValPubKey) -> AccPubKey:
    vals = bech32_decode(data)
    if vals[1] is None:
        raise ValueError(f"invalid bech32: {data}")
    return AccPubKey(bech32_encode("terrapub", vals[1]))


def is_val_pubkey(data: str) -> bool:
    return check_prefix_and_length("terravaloperpub", data, 83)


def to_val_pubkey(data: AccPubKey) -> ValPubKey:
    vals = bech32_decode(data)
    if vals[1] is None:
        raise ValueError(f"invalid bech32: {data}")
    return ValPubKey(bech32_encode("terravaloperpub", vals[1]))


def is_valcons_pubkey(data: str) -> ValConsPubKey:
    return check_prefix_and_length("terravalconspub", data, 83)



from typing import List, Optional

from terra_sdk.core import AccAddress
from terra_sdk.core.msgauth import AuthorizationGrant

from ._base import BaseAPI


class AsyncMsgAuthAPI(BaseAPI):
    async def grants(
        self, granter: AccAddress, grantee: AccAddress, msg_type: Optional[str] = None
    ) -> List[AuthorizationGrant]:
        if msg_type is None:
            res = await self._c._get(
                f"/msgauth/granters/{granter}/grantees/{grantee}/grants"
            )
            return [AuthorizationGrant.from_data(x) for x in res]
        else:
            res = await self._c._get(
                f"/msgauth/granters/{granter}/grantees/{grantee}/grants/{msg_type}"
            )
            if res is None:
                return []
            else:
                return [AuthorizationGrant.from_data(x) for x in res]


class MsgAuthAPI(BaseAPI):
    def grants(
        self, granter: AccAddress, grantee: AccAddress, msg_type: Optional[str] = None
    ) -> List[AuthorizationGrant]:
        """Fetches current active message authorization grants.

        Args:
            granter (AccAddress): granter account address
            grantee (AccAddress): grantee account address
            msg_type (Optional[str], optional): message type. Defaults to None.

        Returns:
            List[AuthorizationGrant]: message authorization grants matching criteria
        """
        if msg_type is None:
            res = self._c._get(f"/msgauth/granters/{granter}/grantees/{grantee}/grants")
            return [AuthorizationGrant.from_data(x) for x in res]
        else:
            res = self._c._get(
                f"/msgauth/granters/{granter}/grantees/{grantee}/grants/{msg_type}"
            )
            if res is None:
                return []
            else:
                return [AuthorizationGrant.from_data(x) for x in res]



from .data import (
    Account,
    LazyGradedVestingAccount,
    StdFee,
    StdSignature,
    StdSignMsg,
    StdTx,
    TxInfo,
    TxLog,
    parse_tx_logs,
)

__all__ = [
    "Account",
    "LazyGradedVestingAccount",
    "StdSignature",
    "StdFee",
    "StdSignMsg",
    "StdTx",
    "TxLog",
    "TxInfo",
    "parse_tx_logs",
]



import asyncio
from typing import Any, Callable

from terra_sdk.client.lcd import AsyncLCDClient

__all__ = ["run_with_async_client"]


async def _run(
    loop: asyncio.AbstractEventLoop, fn: Callable[[AsyncLCDClient], Any], **kwargs
):
    async with AsyncLCDClient(loop=loop, **kwargs) as terra:
        await fn(terra)


def run_with_async_client(fn: Callable[[AsyncLCDClient], Any], **kwargs):
    loop = asyncio.new_event_loop()
    loop.run_until_complete(_run(loop, fn, **kwargs))



from terra_sdk.core import Coin, Dec

from ._base import BaseAPI


class AsyncMarketAPI(BaseAPI):
    def swap_rate(self, offer_coin: Coin, ask_denom: str) -> Coin:
        params = {"offer_coin": str(offer_coin), "ask_denom": ask_denom}
        res = self._c._get("/market/swap", params)
        return Coin.from_data(res)

    def terra_pool_delta(self) -> Dec:
        res = self._c._get("/market/terra_pool_delta")
        return Dec(res)

    def parameters(self) -> dict:
        return self._c._get("/market/parameters")


class MarketAPI(BaseAPI):
    def swap_rate(self, offer_coin: Coin, ask_denom: str) -> Coin:
        """Simulates a swap given an amount offered and a target denom.

        Args:
            offer_coin (Coin): amount offered (swap from)
            ask_denom (str): target denom (swap to)

        Returns:
            Coin: simulated amount received
        """
        params = {"offer_coin": str(offer_coin), "ask_denom": ask_denom}
        res = self._c._get("/market/swap", params)
        return Coin.from_data(res)

    def terra_pool_delta(self) -> Dec:
        """Fetches the Terra pool delta.

        Returns:
            Dec: Terra pool delta
        """
        res = self._c._get("/market/terra_pool_delta")
        return Dec(res)

    def parameters(self) -> dict:
        """Fetches the Market module's parameters.

        Returns:
            dict: Market module parameters
        """
        return self._c._get("/market/parameters")



from __future__ import annotations

from typing import Optional
from urllib.parse import urljoin

import requests

from terra_sdk.core import Coins, Numeric
from terra_sdk.exceptions import LCDResponseError
from terra_sdk.key.key import Key
from terra_sdk.util.json import dict_to_data

from .api.auth import AuthAPI
from .api.bank import BankAPI
from .api.distribution import DistributionAPI
from .api.gov import GovAPI
from .api.market import MarketAPI
from .api.mint import MintAPI
from .api.msgauth import MsgAuthAPI
from .api.oracle import OracleAPI
from .api.slashing import SlashingAPI
from .api.staking import StakingAPI
from .api.supply import SupplyAPI
from .api.tendermint import TendermintAPI
from .api.treasury import TreasuryAPI
from .api.tx import TxAPI
from .api.wasm import WasmAPI
from .wallet import Wallet


class LCDClient:
    def __init__(
        self,
        url: str,
        chain_id: str = None,
        gas_prices: Coins.Input = None,
        gas_adjustment: Numeric.Input = None,
    ):
        self.session = requests.session()
        self.session.headers = {"Accept": "application/json"}  # type: ignore
        self.chain_id = chain_id
        self.url = url
        self.gas_prices = Coins(gas_prices)
        self.gas_adjustment = gas_adjustment
        self._last_request_height = None

        self.auth = AuthAPI(self)
        self.bank = BankAPI(self)
        self.distribution = DistributionAPI(self)
        self.gov = GovAPI(self)
        self.market = MarketAPI(self)
        self.mint = MintAPI(self)
        self.msgauth = MsgAuthAPI(self)
        self.oracle = OracleAPI(self)
        self.slashing = SlashingAPI(self)
        self.staking = StakingAPI(self)
        self.supply = SupplyAPI(self)
        self.tendermint = TendermintAPI(self)
        self.treasury = TreasuryAPI(self)
        self.wasm = WasmAPI(self)
        self.tx = TxAPI(self)

    def wallet(self, key: Key) -> Wallet:
        return Wallet(self, key)

    def _get(self, endpoint: str, params: Optional[dict] = None, raw: bool = False):
        response = self.session.get(urljoin(self.url, endpoint), params=params)
        result = response.json()
        if not str(response.status_code).startswith("2"):
            raise LCDResponseError(message=result.get("error"), response=response)
        try:
            self._last_request_height = result["height"]
        except KeyError:
            self._last_request_height = None
        return result if raw else result["result"]

    def _post(self, endpoint: str, data: Optional[dict] = None, raw: bool = False):
        response = self.session.post(
            urljoin(self.url, endpoint), json=data and dict_to_data(data)
        )
        result = response.json()
        if not str(response.status_code).startswith("2"):
            raise LCDResponseError(message=result.get("error"), response=response)
        try:
            self._last_request_height = result["height"]
        except KeyError:
            self._last_request_height = None
        return result if raw else result["result"]



from .msgs import (
    MsgExecuteContract,
    MsgInstantiateContract,
    MsgMigrateContract,
    MsgStoreCode,
    MsgUpdateContractOwner,
    b64_to_dict,
    dict_to_b64,
)

__all__ = [
    "b64_to_dict",
    "dict_to_b64",
    "MsgStoreCode",
    "MsgInstantiateContract",
    "MsgExecuteContract",
    "MsgMigrateContract",
    "MsgUpdateContractOwner",
]



from __future__ import annotations

import copy

import attr

from terra_sdk.core import Dec, ValAddress, ValConsPubKey
from terra_sdk.util.json import JSONSerializable, dict_to_data

__all__ = [
    "CommissionRates",
    "Commission",
    "Description",
    "Validator",
]


@attr.s
class CommissionRates(JSONSerializable):

    rate: Dec = attr.ib(converter=Dec)
    max_rate: Dec = attr.ib(converter=Dec)
    max_change_rate: Dec = attr.ib(converter=Dec)

    @classmethod
    def from_data(cls, data: dict) -> CommissionRates:
        return cls(
            rate=data["rate"],
            max_rate=data["max_rate"],
            max_change_rate=data["max_change_rate"],
        )


@attr.s
class Commission(JSONSerializable):

    commission_rates: CommissionRates = attr.ib()
    update_time: str = attr.ib()

    @classmethod
    def from_data(cls, data: dict) -> Commission:
        return cls(
            commission_rates=CommissionRates.from_data(data["commission_rates"]),
            update_time=data["update_time"],
        )


# from cosmos


@attr.s
class Description(JSONSerializable):

    DO_NOT_MODIFY = "[do-not-modify]"

    moniker: str = attr.ib()
    identity: str = attr.ib()
    website: str = attr.ib()
    details: str = attr.ib()

    @classmethod
    def from_data(cls, data) -> Description:
        return cls(data["moniker"], data["identity"], data["website"], data["details"])


@attr.s
class Validator(JSONSerializable):

    operator_address: ValAddress = attr.ib()
    consensus_pubkey: ValConsPubKey = attr.ib()
    jailed: bool = attr.ib()
    status: int = attr.ib(converter=int)
    tokens: int = attr.ib(converter=int)
    delegator_shares: Dec = attr.ib(converter=Dec)
    description: Description = attr.ib()
    unbonding_height: int = attr.ib(converter=int)
    unbonding_time: str = attr.ib()
    commission: Commission = attr.ib()
    min_self_delegation: int = attr.ib(converter=int)

    def to_data(self) -> dict:
        d = copy.deepcopy(self.__dict__)
        d["min_self_delegation"] = str(d["min_self_delegation"])
        d["tokens"] = str(d["tokens"])
        d["unbonding_height"] = str(d["unbonding_height"])
        return dict_to_data(d)

    @classmethod
    def from_data(cls, data: dict) -> Validator:
        return cls(
            operator_address=data["operator_address"],
            consensus_pubkey=data["consensus_pubkey"],
            jailed=data["jailed"],
            status=data["status"],
            tokens=data["tokens"],
            delegator_shares=data["delegator_shares"],
            description=Description.from_data(data["description"]),
            unbonding_height=data["unbonding_height"],
            unbonding_time=data["unbonding_time"],
            commission=Commission.from_data(data["commission"]),
            min_self_delegation=data["min_self_delegation"],
        )



import abc
import base64
import hashlib
from typing import Optional

from bech32 import bech32_encode, convertbits

from terra_sdk.core.auth import StdSignature, StdSignMsg, StdTx

BECH32_PUBKEY_DATA_PREFIX = "eb5ae98721"

__all__ = ["Key"]


def get_bech(prefix: str, payload: str) -> str:
    data = convertbits(bytes.fromhex(payload), 8, 5)
    if data is None:
        raise ValueError(f"could not parse data: prefix {prefix}, payload {payload}")
    return bech32_encode(prefix, data)  # base64 -> base32


def address_from_public_key(public_key: bytes) -> bytes:
    sha = hashlib.sha256()
    rip = hashlib.new("ripemd160")
    sha.update(public_key)
    rip.update(sha.digest())
    return rip.digest()


def pubkey_from_public_key(public_key: bytes) -> bytes:
    arr = bytearray.fromhex(BECH32_PUBKEY_DATA_PREFIX)
    arr += bytearray(public_key)
    return bytes(arr)


class Key:

    public_key: Optional[bytes]
    raw_address: Optional[bytes]
    raw_pubkey: Optional[bytes]

    def __init__(self, public_key: Optional[bytes] = None):
        self.public_key = public_key
        if public_key:
            self.raw_address = address_from_public_key(public_key)
            self.raw_pubkey = pubkey_from_public_key(public_key)

    @abc.abstractmethod
    def sign(self, payload: bytes) -> bytes:
        raise NotImplementedError("an instance of Key must implement Key.sign")

    @property
    def acc_address(self) -> str:
        if not self.raw_address:
            raise ValueError("could not compute acc_address: missing raw_address")
        return get_bech("terra", self.raw_address.hex())

    @property
    def val_address(self) -> str:
        if not self.raw_address:
            raise ValueError("could not compute val_address: missing raw_address")
        return get_bech("terravaloper", self.raw_address.hex())

    @property
    def acc_pubkey(self) -> str:
        if not self.raw_pubkey:
            raise ValueError("could not compute acc_pubkey: missing raw_pubkey")
        return get_bech("terrapub", self.raw_pubkey.hex())

    @property
    def val_pubkey(self) -> str:
        if not self.raw_pubkey:
            raise ValueError("could not compute val_pubkey: missing raw_pubkey")
        return get_bech("terravaloperpub", self.raw_pubkey.hex())

    def create_signature(self, tx: StdSignMsg) -> StdSignature:
        if self.public_key is None:
            raise ValueError(
                "signature could not be created: Key instance missing public_key"
            )

        sig_buffer = self.sign(tx.to_json().strip().encode())
        return StdSignature.from_data(
            {
                "signature": base64.b64encode(sig_buffer).decode(),
                "pub_key": {
                    "type": "tendermint/PubKeySecp256k1",
                    "value": base64.b64encode(self.public_key).decode(),
                },
            }
        )

    def sign_tx(self, tx: StdSignMsg) -> StdTx:
        sig = self.create_signature(tx)
        return StdTx(tx.msgs, tx.fee, [sig], tx.memo)



from __future__ import annotations

import copy

import attr

from terra_sdk.core import Coins
from terra_sdk.util.base import BaseTerraData
from terra_sdk.util.json import JSONSerializable, dict_to_data

__all__ = ["Proposal", "Content"]


@attr.s
class Content(BaseTerraData):

    title: str = attr.ib()
    description: str = attr.ib()

    @staticmethod
    def from_data(data: dict) -> Content:
        from terra_sdk.util.parse_content import parse_content

        return parse_content(data)


@attr.s
class Proposal(JSONSerializable):

    id: int = attr.ib(converter=int)
    content: Content = attr.ib()
    proposal_status: str = attr.ib()
    final_tally_result: dict = attr.ib()
    submit_time: str = attr.ib()
    deposit_end_time: str = attr.ib()
    total_deposit: Coins = attr.ib(converter=Coins)
    voting_start_time: str = attr.ib()
    voting_end_time: str = attr.ib()

    def to_data(self) -> dict:
        d = copy.deepcopy(self.__dict__)
        d["id"] = str(d["id"])
        return dict_to_data(d)

    @classmethod
    def from_data(cls, data: dict) -> Proposal:
        return cls(
            id=data["id"],
            content=Content.from_data(data["content"]),
            proposal_status=data["proposal_status"],
            final_tally_result=data["final_tally_result"],
            submit_time=data["submit_time"],
            deposit_end_time=data["deposit_end_time"],
            total_deposit=Coins.from_data(data["total_deposit"]),
            voting_start_time=data["voting_start_time"],
            voting_end_time=data["voting_end_time"],
        )



from __future__ import annotations

import attr

from terra_sdk.core import AccAddress, Coin
from terra_sdk.core.msg import Msg

__all__ = ["MsgSwap", "MsgSwapSend"]


@attr.s
class MsgSwap(Msg):

    type = "market/MsgSwap"
    action = "swap"

    trader: AccAddress = attr.ib()
    offer_coin: Coin = attr.ib(converter=Coin.parse)  # type: ignore
    ask_denom: str = attr.ib()

    @classmethod
    def from_data(cls, data: dict) -> MsgSwap:
        data = data["value"]
        return cls(
            trader=data["trader"],
            offer_coin=Coin.from_data(data["offer_coin"]),
            ask_denom=data["ask_denom"],
        )


@attr.s
class MsgSwapSend(Msg):

    type = "market/MsgSwapSend"
    action = "swapsend"

    from_address: AccAddress = attr.ib()
    to_address: AccAddress = attr.ib()
    offer_coin: Coin = attr.ib(converter=Coin.parse)  # type: ignore
    ask_denom: str = attr.ib()

    @classmethod
    def from_data(cls, data: dict) -> MsgSwapSend:
        data = data["value"]
        return cls(
            from_address=data["from_address"],
            to_address=data["to_address"],
            offer_coin=Coin.from_data(data["offer_coin"]),
            ask_denom=data["ask_denom"],
        )



from __future__ import annotations

from terra_sdk.core import AccAddress, Coins
from terra_sdk.core.msg import Msg

__all__ = ["MsgSend", "MsgMultiSend"]

import attr


@attr.s
class MsgSend(Msg):

    type = "bank/MsgSend"
    action = "send"

    from_address: AccAddress = attr.ib()
    to_address: AccAddress = attr.ib()
    amount: Coins = attr.ib(converter=Coins)

    @classmethod
    def from_data(cls, data: dict) -> MsgSend:
        data = data["value"]
        return cls(
            from_address=data["from_address"],
            to_address=data["to_address"],
            amount=Coins.from_data(data["amount"]),
        )


@attr.s
class MsgMultiSend(Msg):

    type = "bank/MsgMultiSend"
    action = "multisend"

    # TODO: improve interface - match terra.js
    inputs = attr.ib()
    outputs = attr.ib()

    @classmethod
    def from_data(cls, data: dict) -> MsgMultiSend:
        data = data["value"]
        return cls(
            inputs=data["inputs"],
            outputs=data["outputs"],
        )



from typing import List, Optional, Union

from ._base import BaseAPI

__all__ = ["TxAPI"]

from terra_sdk.core import AccAddress, Coin, Coins, Numeric
from terra_sdk.core.auth import StdFee, StdSignMsg, StdTx, TxInfo
from terra_sdk.core.broadcast import (
    AsyncTxBroadcastResult,
    BlockTxBroadcastResult,
    SyncTxBroadcastResult,
)
from terra_sdk.core.msg import Msg
from terra_sdk.util.hash import hash_amino


class AsyncTxAPI(BaseAPI):
    async def tx_info(self, tx_hash: str) -> TxInfo:
        return TxInfo.from_data(await self._c._get(f"/txs/{tx_hash}", raw=True))

    async def create(
        self,
        source_address: AccAddress,
        msgs: List[Msg],
        fee: Optional[StdFee] = None,
        memo: str = "",
        gas_prices: Optional[Coins.Input] = None,
        gas_adjustment: Optional[Numeric.Input] = None,
        denoms: Optional[List[str]] = None,
        account_number: Optional[int] = None,
        sequence: Optional[int] = None,
    ) -> StdSignMsg:
        if fee is None:
            # create the fake fee
            balance = await self._c.bank.balance(source_address)
            balance_one = [Coin(c.denom, 1) for c in balance]

            # estimate the fee
            tx = StdTx(msgs, StdFee(0, balance_one), [], memo)
            fee = await self.estimate_fee(tx, gas_prices, gas_adjustment, denoms)

        if account_number is None or sequence is None:
            account = await self._c.auth.account_info(source_address)
            if account_number is None:
                account_number = account.account_number
            if sequence is None:
                sequence = account.sequence

        return StdSignMsg(
            self._c.chain_id, account_number or 0, sequence or 0, fee, msgs, memo
        )

    async def estimate_fee(
        self,
        tx: Union[StdSignMsg, StdTx],
        gas_prices: Optional[Coins.Input] = None,
        gas_adjustment: Optional[Numeric.Input] = None,
        denoms: Optional[List[str]] = None,
    ) -> StdFee:
        gas_prices = gas_prices or self._c.gas_prices
        gas_adjustment = gas_adjustment or self._c.gas_adjustment

        if isinstance(tx, StdSignMsg):
            tx_value = tx.to_stdtx().to_data()["value"]
        else:
            tx_value = tx.to_data()["value"]

        tx_value["fee"]["gas"] = "0"

        data = {
            "tx": tx_value,
            "gas_prices": gas_prices and Coins(gas_prices).to_data(),
            "gas_adjustment": gas_adjustment and str(gas_adjustment),
        }

        res = await self._c._post("/txs/estimate_fee", data)
        fees = Coins.from_data(res["fees"])
        # only pick the denoms we are interested in?
        if denoms:
            fees = fees.filter(lambda c: c.denom in denoms)  # type: ignore
        return StdFee(int(res["gas"]), fees)

    async def encode(self, tx: StdTx) -> str:
        res = await self._c._post("/txs/encode", tx.to_data())
        return res["tx"]

    async def hash(self, tx: StdTx) -> str:
        amino = await self.encode(tx)
        return hash_amino(amino)

    async def _broadcast(self, tx: StdTx, mode: str) -> dict:
        data = {"tx": tx.to_data()["value"], "mode": mode}
        return await self._c._post("/txs", data, raw=True)

    async def broadcast_sync(self, tx: StdTx) -> SyncTxBroadcastResult:
        res = await self._broadcast(tx, "sync")
        return SyncTxBroadcastResult(
            height=res["height"],
            txhash=res["txhash"],
            raw_log=res["raw_log"],
            code=res.get("code"),
            codespace=res.get("codespace"),
        )

    async def broadcast_async(self, tx: StdTx) -> AsyncTxBroadcastResult:
        res = await self._broadcast(tx, "async")
        return AsyncTxBroadcastResult(
            height=res["height"],
            txhash=res["txhash"],
        )

    async def broadcast(self, tx: StdTx) -> BlockTxBroadcastResult:
        res = await self._broadcast(tx, "block")
        return BlockTxBroadcastResult(
            height=res["height"],
            txhash=res["txhash"],
            raw_log=res["raw_log"],
            gas_wanted=res["gas_wanted"],
            gas_used=res["gas_used"],
            logs=res.get("logs"),
            code=res.get("code"),
            codespace=res.get("codespace"),
        )

    async def search(self, options: dict = {}) -> dict:
        res = await self._c._get("/txs", options, raw=True)
        return res


class TxAPI(BaseAPI):
    def tx_info(self, tx_hash: str) -> TxInfo:
        """Fetches information for an included transaction given a tx hash.

        Args:
            tx_hash (str): hash of transaction to lookup

        Returns:
            TxInfo: transaction info
        """
        return TxInfo.from_data(self._c._get(f"/txs/{tx_hash}", raw=True))

    def create(
        self,
        source_address: AccAddress,
        msgs: List[Msg],
        fee: Optional[StdFee] = None,
        memo: str = "",
        gas_prices: Optional[Coins.Input] = None,
        gas_adjustment: Optional[Numeric.Input] = None,
        denoms: Optional[List[str]] = None,
        account_number: Optional[int] = None,
        sequence: Optional[int] = None,
    ) -> StdSignMsg:
        """Create a new unsigned transaction, with helpful utilities such as lookup of
        chain ID, account number, sequence and fee estimation.

        Args:
            source_address (AccAddress): transaction sender's account address
            msgs (List[Msg]): list of messages to include
            fee (Optional[StdFee], optional): fee to use (estimates if empty). Defaults to None.
            memo (str, optional): memo to use. Defaults to "".
            gas_prices (Optional[Coins.Input], optional): gas prices for fee estimation. Defaults to None.
            gas_adjustment (Optional[Numeric.Input], optional): gas adjustment for fee estimation. Defaults to None.
            denoms (Optional[List[str]], optional): list of denoms to use for gas fee. Defaults to None.
            account_number (Optional[int], optional): account number to use. Defaults to None.
            sequence (Optional[int], optional): sequence number to use. Defaults to None.

        Returns:
            StdSignMsg: unsigned tx
        """
        if fee is None:
            # create the fake fee
            balance = self._c.bank.balance(source_address)
            balance_one = [Coin(c.denom, 1) for c in balance]

            # estimate the fee
            tx = StdTx(msgs, StdFee(0, balance_one), [], memo)
            fee = self.estimate_fee(tx, gas_prices, gas_adjustment, denoms)

        if account_number is None or sequence is None:
            account = self._c.auth.account_info(source_address)
            if account_number is None:
                account_number = account.account_number
            if sequence is None:
                sequence = account.sequence

        return StdSignMsg(
            self._c.chain_id, account_number or 0, sequence or 0, fee, msgs, memo
        )

    def estimate_fee(
        self,
        tx: Union[StdSignMsg, StdTx],
        gas_prices: Optional[Coins.Input] = None,
        gas_adjustment: Optional[Numeric.Input] = None,
        denoms: Optional[List[str]] = None,
    ) -> StdFee:
        """Estimates the proper fee to apply by simulating it within the node.

        Args:
            tx (Union[StdSignMsg, StdTx]): transaction to estimate fee for
            gas_prices (Optional[Coins.Input], optional): gas prices to use. Defaults to None.
            gas_adjustment (Optional[Numeric.Input], optional): gas adjustment to use. Defaults to None.
            denoms (Optional[List[str]], optional): list of denoms to use to pay for gas. Defaults to None.

        Returns:
            StdFee: estimated fee
        """
        gas_prices = gas_prices or self._c.gas_prices
        gas_adjustment = gas_adjustment or self._c.gas_adjustment

        if isinstance(tx, StdSignMsg):
            tx_value = tx.to_stdtx().to_data()["value"]
        else:
            tx_value = tx.to_data()["value"]

        tx_value["fee"]["gas"] = "0"

        data = {
            "tx": tx_value,
            "gas_prices": gas_prices and Coins(gas_prices).to_data(),
            "gas_adjustment": gas_adjustment and str(gas_adjustment),
        }

        res = self._c._post("/txs/estimate_fee", data)
        fees = Coins.from_data(res["fees"])
        # only pick the denoms we are interested in?
        if denoms:
            fees = fees.filter(lambda c: c.denom in denoms)  # type: ignore
        return StdFee(int(res["gas"]), fees)

    def encode(self, tx: StdTx) -> str:
        """Fetches a transaction's amino encoding.

        Args:
            tx (StdTx): transaction to encode

        Returns:
            str: base64 string containing amino-encoded tx
        """
        res = self._c._post("/txs/encode", tx.to_data())
        return res["tx"]

    def hash(self, tx: StdTx) -> str:
        """Compute hash for a transaction.

        Args:
            tx (StdTx): transaction to hash

        Returns:
            str: transaction hash
        """
        amino = self.encode(tx)
        return hash_amino(amino)

    def _broadcast(self, tx: StdTx, mode: str) -> dict:
        data = {"tx": tx.to_data()["value"], "mode": mode}
        return self._c._post("/txs", data, raw=True)

    def broadcast_sync(self, tx: StdTx) -> SyncTxBroadcastResult:
        """Broadcasts a transaction using the ``sync`` broadcast mode.

        Args:
            tx (StdTx): transaction to broadcast

        Returns:
            SyncTxBroadcastResult: result
        """
        res = self._broadcast(tx, "sync")
        return SyncTxBroadcastResult(
            height=res["height"],
            txhash=res["txhash"],
            raw_log=res["raw_log"],
            code=res.get("code"),
            codespace=res.get("codespace"),
        )

    def broadcast_async(self, tx: StdTx) -> AsyncTxBroadcastResult:
        """Broadcasts a transaction using the ``async`` broadcast mode.

        Args:
            tx (StdTx): transaction to broadcast

        Returns:
            AsyncTxBroadcastResult: result
        """
        res = self._broadcast(tx, "async")
        return AsyncTxBroadcastResult(
            height=res["height"],
            txhash=res["txhash"],
        )

    def broadcast(self, tx: StdTx) -> BlockTxBroadcastResult:
        """Broadcasts a transaction using the ``block`` broadcast mode.

        Args:
            tx (StdTx): transaction to broadcast

        Returns:
            BlockTxBroadcastResult: result
        """
        res = self._broadcast(tx, "block")
        return BlockTxBroadcastResult(
            height=res["height"],
            txhash=res["txhash"],
            raw_log=res["raw_log"],
            gas_wanted=res["gas_wanted"],
            gas_used=res["gas_used"],
            logs=res.get("logs"),
            code=res.get("code"),
            codespace=res.get("codespace"),
        )

    def search(self, options: dict = {}) -> dict:
        """Searches for transactions given critera.

        Args:
            options (dict, optional): dictionary containing options. Defaults to {}.

        Returns:
            dict: transaction search results
        """
        res = self._c._get("/txs", options, raw=True)
        return res



from typing import List, Optional

from terra_sdk.core import ValConsPubKey

from ._base import BaseAPI


class AsyncSlashingAPI(BaseAPI):
    async def signing_infos(
        self, val_cons_pub_key: Optional[ValConsPubKey] = None
    ) -> List[dict]:
        if val_cons_pub_key is None:
            url = "/slashing/signing_infos"
        else:
            url = f"/slashing/validators/{val_cons_pub_key}/signing_info"
        return await self._c._get(url)

    async def parameters(self) -> dict:
        return await self._c._get("/slashing/parameters")


class SlashingAPI(BaseAPI):
    def signing_infos(
        self, val_cons_pub_key: Optional[ValConsPubKey] = None
    ) -> List[dict]:
        """Fetches signing infos, optionally filtering by validator consensus public key.

        Args:
            val_cons_pub_key (Optional[ValConsPubKey], optional): validator consensus public key. Defaults to None.

        Returns:
            List[dict]: signing infos
        """
        if val_cons_pub_key is None:
            url = "/slashing/signing_infos"
        else:
            url = f"/slashing/validators/{val_cons_pub_key}/signing_info"
        return self._c._get(url)

    def parameters(self) -> dict:
        """Fetches Slashing module parameters.

        Returns:
            dict: Slashing module parameters
        """
        return self._c._get("/slashing/parameters")



from __future__ import annotations

from terra_sdk.core.auth import StdSignMsg, StdTx
from terra_sdk.key.key import Key

__all__ = ["Wallet", "AsyncWallet"]


class AsyncWallet:
    def __init__(self, lcd, key: Key):
        self.lcd = lcd
        self.key = key

    async def account_number(self) -> int:
        res = await self.lcd.auth.account_info(self.key.acc_address)
        return res.account_number

    async def sequence(self) -> int:
        res = await self.lcd.auth.account_info(self.key.acc_address)
        return res.sequence

    async def account_number_and_sequence(self) -> dict:
        res = await self.lcd.auth.account_info(self.key.acc_address)
        return {"account_number": res.account_number, "sequence": res.sequence}

    async def create_tx(self, *args, **kwargs) -> StdSignMsg:
        return await self.lcd.tx.create(self.key.acc_address, *args, **kwargs)

    async def create_and_sign_tx(self, *args, **kwargs) -> StdTx:
        return self.key.sign_tx(await self.create_tx(*args, **kwargs))


class Wallet:
    def __init__(self, lcd, key: Key):
        self.lcd = lcd
        self.key = key

    def account_number(self) -> int:
        res = self.lcd.auth.account_info(self.key.acc_address)
        return res.account_number

    def sequence(self) -> int:
        res = self.lcd.auth.account_info(self.key.acc_address)
        return res.sequence

    def account_number_and_sequence(self) -> dict:
        res = self.lcd.auth.account_info(self.key.acc_address)
        return {"account_number": res.account_number, "sequence": res.sequence}

    def create_tx(self, *args, **kwargs) -> StdSignMsg:
        return self.lcd.tx.create(self.key.acc_address, *args, **kwargs)

    def create_and_sign_tx(self, *args, **kwargs) -> StdTx:
        return self.key.sign_tx(self.create_tx(*args, **kwargs))



from __future__ import annotations

from .data import Content

__all__ = ["TextProposal"]


class TextProposal(Content):

    type = "gov/TextProposal"

    @classmethod
    def from_data(cls, data: dict) -> TextProposal:
        data = data["value"]
        return cls(title=data["title"], description=data["description"])



import json
from typing import Any

from ._base import BaseAPI


class AsyncWasmAPI(BaseAPI):
    async def code_info(self, code_id: int) -> dict:
        return await self._c._get(f"/wasm/codes/{code_id}")

    async def contract_info(self, contract_address: str) -> dict:
        res = await self._c._get(f"/wasm/contracts/{contract_address}")
        return res

    async def contract_query(self, contract_address: str, query: dict) -> Any:
        params = {"query_msg": json.dumps(query)}
        return await self._c._get(f"/wasm/contracts/{contract_address}/store", params)

    async def parameters(self) -> dict:
        return await self._c._get("/wasm/parameters")


class WasmAPI(BaseAPI):
    def code_info(self, code_id: int) -> dict:
        """Fetches information about an uploaded code.

        Args:
            code_id (int): code ID

        Returns:
            dict: code information
        """
        return self._c._get(f"/wasm/codes/{code_id}")

    def contract_info(self, contract_address: str) -> dict:
        """Fetches information about an instantiated contract.

        Args:
            contract_address (str): contract address

        Returns:
            dict: contract information
        """
        res = self._c._get(f"/wasm/contracts/{contract_address}")
        return res

    def contract_query(self, contract_address: str, query_msg: dict) -> Any:
        """Runs a QueryMsg on a contract.

        Args:
            contract_address (str): contract address
            query_msg (dict): QueryMsg to run

        Returns:
            Any: results of query
        """
        params = {"query_msg": json.dumps(query_msg)}
        return self._c._get(f"/wasm/contracts/{contract_address}/store", params)

    def parameters(self) -> dict:
        """Fetches the Wasm module parameters.

        Returns:
            dict: Wasm module parameters
        """
        return self._c._get("/wasm/parameters")



from terra_sdk.core import AccAddress, Coins

from ._base import BaseAPI


class AsyncBankAPI(BaseAPI):
    async def balance(self, address: AccAddress) -> Coins:
        res = await self._c._get(f"/bank/balances/{address}")
        return Coins.from_data(res)


class BankAPI(BaseAPI):
    def balance(self, address: AccAddress) -> Coins:
        """Fetches the current balance of an account.

        Args:
            address (AccAddress): account address

        Returns:
            Coins: balance
        """
        res = self._c._get(f"/bank/balances/{address}")
        return Coins.from_data(res)



from __future__ import annotations

import copy
from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional, Union

from .coin import Coin
from .numeric import Numeric


class Coins:

    Input = Union[Iterable[Coin], str, Dict[str, Numeric.Input], Dict[str, Coin]]

    _coins: Dict[str, Coin]

    def __repr__(self) -> str:
        if len(self) == 0:
            return "Coins()"
        else:
            return f'Coins("{self!s}")'

    def __str__(self) -> str:
        return ",".join(str(coin) for coin in self)

    @classmethod
    def from_str(cls, s: str) -> Coins:
        coin_strings = s.split(r",")
        return Coins(Coin.from_str(cs) for cs in coin_strings)

    def __init__(self, arg: Optional[Coins.Input] = {}, **denoms):

        if arg is None:
            self._coins = {}
            return

        # arg should be an iterable
        try:
            iter(arg)
        except TypeError:
            raise TypeError(f"could not create Coins object with argument: {arg!s}")

        if isinstance(arg, Coins):
            self._coins = copy.deepcopy(arg._coins)
            return

        if isinstance(arg, str):
            self._coins = Coins.from_str(arg)._coins
            return

        self._coins = Coins(denoms)._coins if denoms else {}

        coins: Iterable[Coin]
        if isinstance(arg, dict):
            coins = [Coin(denom, arg[denom]) for denom in arg]
        else:
            coins = arg
        for coin in coins:
            x = self._coins.get(coin.denom)
            if x is not None:
                self._coins[coin.denom] = x + coin
            else:
                self._coins[coin.denom] = coin

        # check homogeneous
        if not all([c.is_dec_coin() for c in self]) and not all(
            [c.is_int_coin() for c in self]
        ):
            raise TypeError(
                f"non-homogenous instantiation of Coins not supported: {self!s}"
            )

    def __getitem__(self, denom: str) -> Coin:
        return self._coins[denom]

    def get(self, denom: str) -> Optional[Coin]:
        return self._coins.get(denom)

    @classmethod
    def from_data(cls, data: list) -> Coins:
        coins = map(Coin.from_data, data)
        return cls(coins)

    def to_data(self) -> List[dict]:
        return [coin.to_data() for coin in self]

    def denoms(self) -> List[str]:
        return [c.denom for c in self]

    def to_dec_coins(self) -> Coins:
        return Coins(c.to_dec_coin() for c in self)

    def to_int_coins(self) -> Coins:
        return Coins(c.to_int_coin() for c in self)

    def add(self, other: Union[Coin, Coins]) -> Coins:
        if isinstance(other, Coin):
            return Coins([other, *self.to_list()])
        else:
            return Coins([*other.to_list(), *self.to_list()])

    def __add__(self, other: Union[Coin, Coins]) -> Coins:
        return self.add(other)

    def sub(self, other: Union[Coin, Coins]) -> Coins:
        return self.add(self.mul(-1))

    def __sub__(self, other: Union[Coin, Coins]) -> Coins:
        return self.sub(other)

    def mul(self, other: Numeric.Input) -> Coins:
        return Coins(coin.mul(other) for coin in self)

    def __mul__(self, other: Numeric.Input) -> Coins:
        return self.mul(other)

    def div(self, other: Numeric.Input) -> Coins:
        return Coins(coin.div(other) for coin in self)

    def __truediv__(self, other: Numeric.Input) -> Coins:
        return Coins(coin / other for coin in self)

    def __floordiv__(self, other: Numeric.Input) -> Coins:
        return Coins(coin // other for coin in self)

    def to_list(self) -> List[Coin]:
        return sorted(self._coins.values(), key=lambda c: c.denom)

    def filter(self, predicate: Callable[[Coin], bool]) -> Coins:
        return Coins(c for c in self if predicate(c))

    def map(self, fn: Callable[[Coin], Any]) -> Iterator[Any]:
        return map(fn, self)

    def __eq__(self, other) -> bool:
        try:
            return self.to_list() == other.to_list()
        except AttributeError:
            return False

    def __iter__(self):
        return iter(self.to_list())

    def __len__(self):
        return len(self.to_list())

    def __contains__(self, denom: str) -> bool:
        return denom in self._coins



from __future__ import annotations

import attr

from terra_sdk.core import AccAddress, Coins
from terra_sdk.core.gov import Content

__all__ = ["CommunityPoolSpendProposal"]


@attr.s
class CommunityPoolSpendProposal(Content):

    type = "distribution/CommunityPoolSpendProposal"

    title: str = attr.ib()
    description: str = attr.ib()
    recipient: AccAddress = attr.ib()
    amount: Coins = attr.ib(converter=Coins)

    @classmethod
    def from_data(cls, data: dict) -> CommunityPoolSpendProposal:
        data = data["value"]
        return cls(
            title=data["title"],
            description=data["description"],
            recipient=data["recipient"],
            amount=Coins.from_data(data["amount"]),
        )



import hashlib

from ecdsa import SECP256k1, SigningKey
from ecdsa.util import sigencode_string_canonize

from .key import Key

__all__ = ["RawKey"]


def compute_public_key(private_key: bytes) -> bytes:
    return (
        SigningKey.from_string(private_key, curve=SECP256k1)
        .get_verifying_key()
        .to_string("compressed")
    )


class RawKey(Key):

    private_key: bytes

    @classmethod
    def from_hex(cls, private_key_hex: str):
        return cls(bytes.fromhex(private_key_hex))

    def __init__(self, private_key: bytes):
        public_key = compute_public_key(private_key)
        super().__init__(public_key)
        self.private_key = private_key

    def sign(self, payload: bytes) -> bytes:
        sk = SigningKey.from_string(self.private_key, curve=SECP256k1)
        return sk.sign_deterministic(
            payload,
            hashfunc=hashlib.sha256,
            sigencode=sigencode_string_canonize,
        )



from typing import Union

from terra_sdk.core import AccAddress
from terra_sdk.core.auth import Account, LazyGradedVestingAccount

from ._base import BaseAPI


class AsyncAuthAPI(BaseAPI):
    async def account_info(
        self, address: AccAddress
    ) -> Union[Account, LazyGradedVestingAccount]:
        result = await self._c._get(f"/auth/accounts/{address}")
        if result["type"] == "core/Account":
            return Account.from_data(result)
        else:
            return LazyGradedVestingAccount.from_data(result)


class AuthAPI(BaseAPI):
    def account_info(
        self, address: AccAddress
    ) -> Union[Account, LazyGradedVestingAccount]:
        """Fetches the account information.

        Args:
            address (AccAddress): account address

        Returns:
            Union[Account, LazyGradedVestingAccount]: account information
        """
        result = self._c._get(f"/auth/accounts/{address}")
        if result["type"] == "core/Account":
            return Account.from_data(result)
        else:
            return LazyGradedVestingAccount.from_data(result)



from __future__ import annotations

from terra_sdk.util.base import BaseTerraData


class Msg(BaseTerraData):
    @staticmethod
    def from_data(data: dict) -> Msg:
        from terra_sdk.util.parse_msg import parse_msg

        return parse_msg(data)



from __future__ import annotations

from typing import Dict, List, Optional

import attr

from terra_sdk.core.coins import Coins
from terra_sdk.core.msg import Msg
from terra_sdk.core.public_key import PublicKey
from terra_sdk.util.json import JSONSerializable

__all__ = [
    "StdSignature",
    "StdFee",
    "StdSignMsg",
    "StdTx",
    "TxLog",
    "TxInfo",
    "parse_tx_logs",
]


@attr.s
class StdSignature(JSONSerializable):

    signature: str = attr.ib()
    pub_key: Optional[PublicKey] = attr.ib()

    @classmethod
    def from_data(cls, data: dict) -> StdSignature:
        return cls(
            signature=data["signature"],
            pub_key=data.get("pub_key") and PublicKey.from_data(data["pub_key"]),
        )


@attr.s
class StdFee(JSONSerializable):

    gas: int = attr.ib(converter=int)
    amount: Coins = attr.ib(converter=Coins)

    @classmethod
    def from_data(cls, data: dict) -> StdFee:
        return cls(int(data["gas"]), Coins.from_data(data["amount"]))

    def to_data(self) -> dict:
        return {"gas": str(self.gas), "amount": self.amount.to_data()}

    @property
    def gas_prices(self) -> Coins:
        return self.amount.to_dec_coins().div(self.gas)


@attr.s
class StdSignMsg(JSONSerializable):

    chain_id: str = attr.ib()
    account_number: int = attr.ib(converter=int)
    sequence: int = attr.ib(converter=int)
    fee: StdFee = attr.ib()
    msgs: List[Msg] = attr.ib()
    memo: str = attr.ib()

    def to_stdtx(self) -> StdTx:
        return StdTx(self.msgs, self.fee, [], self.memo)

    def to_data(self) -> dict:
        return {
            "chain_id": self.chain_id,
            "account_number": str(self.account_number),
            "sequence": str(self.sequence),
            "fee": self.fee.to_data(),
            "msgs": [m.to_data() for m in self.msgs],
            "memo": self.memo,
        }

    @classmethod
    def from_data(cls, data: dict) -> StdSignMsg:
        return cls(
            data["chain_id"],
            int(data["account_number"]),
            int(data["sequence"]),
            StdFee.from_data(data["fee"]),
            [Msg.from_data(m) for m in data["msgs"]],
            data["memo"],
        )


@attr.s
class StdTx(JSONSerializable):

    msg: List[Msg] = attr.ib()
    fee: StdFee = attr.ib()
    signatures: List[StdSignature] = attr.ib()
    memo: str = attr.ib()

    def to_data(self) -> dict:
        return {
            "type": "core/StdTx",
            "value": {
                "msg": [m.to_data() for m in self.msg],
                "fee": self.fee.to_data(),
                "signatures": [s.to_data() for s in self.signatures],
                "memo": self.memo,
            },
        }

    @classmethod
    def from_data(cls, data: dict) -> StdTx:
        data = data["value"]
        return cls(
            [Msg.from_data(m) for m in data["msg"]],
            StdFee.from_data(data["fee"]),
            [StdSignature.from_data(s) for s in data["signatures"]],
            data["memo"],
        )


def parse_events_by_type(event_data: List[dict]) -> Dict[str, Dict[str, List[str]]]:
    events: Dict[str, Dict[str, List[str]]] = {}
    for ev in event_data:
        for att in ev["attributes"]:
            if ev["type"] not in events:
                events[ev["type"]] = {}
            if att["key"] not in events[ev["type"]]:
                events[ev["type"]][att["key"]] = []
            events[ev["type"]][att["key"]].append(att["value"])
    return events


@attr.s
class TxLog(JSONSerializable):

    msg_index: int = attr.ib(converter=int)
    log: str = attr.ib()
    events: List[dict] = attr.ib()
    events_by_type: Dict[str, Dict[str, List[str]]] = attr.ib(init=False)

    def __attrs_post_init__(self):
        self.events_by_type = parse_events_by_type(self.events)


def parse_tx_logs(logs) -> Optional[List[TxLog]]:
    return (
        [
            TxLog(msg_index=log["msg_index"], log=log["log"], events=log["events"])
            for log in logs
        ]
        if logs
        else None
    )


@attr.s
class TxInfo(JSONSerializable):

    height: int = attr.ib(converter=int)
    txhash: str = attr.ib()
    rawlog: str = attr.ib()
    logs: Optional[List[TxLog]] = attr.ib()
    gas_wanted: int = attr.ib(converter=int)
    gas_used: int = attr.ib(converter=int)
    tx: StdTx = attr.ib()
    timestamp: str = attr.ib()
    code: Optional[int] = attr.ib(default=None)
    codespace: Optional[str] = attr.ib(default=None)

    def to_data(self) -> dict:
        data = {
            "height": str(self.height),
            "txhash": self.txhash,
            "raw_log": self.rawlog,
            "logs": [log.to_data() for log in self.logs] if self.logs else None,
            "gas_wanted": str(self.gas_wanted),
            "gas_used": str(self.gas_used),
            "timestamp": self.timestamp,
            "tx": self.tx.to_data(),
            "code": self.code,
            "codespace": self.codespace,
        }

        if not self.logs:
            del data["logs"]

        if not self.code:
            del data["code"]

        if not self.codespace:
            del data["codespace"]

        return data

    @classmethod
    def from_data(cls, data: dict) -> TxInfo:
        return cls(
            data["height"],
            data["txhash"],
            data["raw_log"],
            parse_tx_logs(data["logs"]),
            data["gas_wanted"],
            data["gas_used"],
            StdTx.from_data(data["tx"]),
            data["timestamp"],
            data.get("code"),
            data.get("codespace"),
        )



from __future__ import annotations

from asyncio import AbstractEventLoop, get_event_loop
from typing import Optional
from urllib.parse import urljoin

from aiohttp import ClientSession

from terra_sdk.core import Coins, Numeric
from terra_sdk.exceptions import LCDResponseError
from terra_sdk.key.key import Key
from terra_sdk.util.json import dict_to_data

from .api.auth import AsyncAuthAPI
from .api.bank import AsyncBankAPI
from .api.distribution import AsyncDistributionAPI
from .api.gov import AsyncGovAPI
from .api.market import AsyncMarketAPI
from .api.mint import AsyncMintAPI
from .api.msgauth import AsyncMsgAuthAPI
from .api.oracle import AsyncOracleAPI
from .api.slashing import AsyncSlashingAPI
from .api.staking import AsyncStakingAPI
from .api.supply import AsyncSupplyAPI
from .api.tendermint import AsyncTendermintAPI
from .api.treasury import AsyncTreasuryAPI
from .api.tx import AsyncTxAPI
from .api.wasm import AsyncWasmAPI
from .wallet import AsyncWallet


class AsyncLCDClient:
    def __init__(
        self,
        url: str,
        chain_id: str = None,
        gas_prices: Coins.Input = None,
        gas_adjustment: Numeric.Input = None,
        loop: Optional[AbstractEventLoop] = None,
    ):

        if loop is None:
            loop = get_event_loop()

        self.session = ClientSession(headers={"Accept": "application/json"}, loop=loop)
        self.chain_id = chain_id
        self.url = url
        self.gas_prices = Coins(gas_prices)
        self.gas_adjustment = gas_adjustment
        self._last_request_height = None

        self.auth = AsyncAuthAPI(self)
        self.bank = AsyncBankAPI(self)
        self.distribution = AsyncDistributionAPI(self)
        self.gov = AsyncGovAPI(self)
        self.market = AsyncMarketAPI(self)
        self.mint = AsyncMintAPI(self)
        self.msgauth = AsyncMsgAuthAPI(self)
        self.oracle = AsyncOracleAPI(self)
        self.slashing = AsyncSlashingAPI(self)
        self.staking = AsyncStakingAPI(self)
        self.supply = AsyncSupplyAPI(self)
        self.tendermint = AsyncTendermintAPI(self)
        self.treasury = AsyncTreasuryAPI(self)
        self.wasm = AsyncWasmAPI(self)
        self.tx = AsyncTxAPI(self)

    def wallet(self, key: Key) -> AsyncWallet:
        return AsyncWallet(self, key)

    async def _get(
        self, endpoint: str, params: Optional[dict] = None, raw: bool = False
    ):
        async with self.session.get(
            urljoin(self.url, endpoint), params=params
        ) as response:
            result = await response.json()
            if not str(response.status).startswith("2"):
                raise LCDResponseError(message=result.get("error"), response=response)
        try:
            self._last_request_height = result["height"]
        except KeyError:
            self._last_request_height = None
        return result if raw else result["result"]

    async def _post(
        self, endpoint: str, data: Optional[dict] = None, raw: bool = False
    ):
        async with self.session.post(
            urljoin(self.url, endpoint), json=data and dict_to_data(data)
        ) as response:
            result = await response.json()
            if not str(response.status).startswith("2"):
                raise LCDResponseError(message=result.get("error"), response=response)
        try:
            self._last_request_height = result["height"]
        except KeyError:
            self._last_request_height = None
        return result if raw else result["result"]

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.session.close()



from __future__ import annotations

import re
from typing import Union

import attr

from terra_sdk.util.json import JSONSerializable

from .numeric import Dec, Numeric


@attr.s(frozen=True)
class Coin(JSONSerializable):

    denom: str = attr.ib()
    amount: Numeric.Output = attr.ib(converter=Numeric.parse)  # type: ignore

    @staticmethod
    def parse(arg: Union[Coin, str, dict]) -> Coin:
        if isinstance(arg, Coin):
            return arg
        elif isinstance(arg, str):
            return Coin.from_str(arg)
        else:
            return Coin.from_data(arg)

    def is_int_coin(self) -> bool:
        return isinstance(self.amount, int)

    def is_dec_coin(self) -> bool:
        return isinstance(self.amount, Dec)

    def to_int_coin(self) -> Coin:
        return Coin(self.denom, int(self.amount))

    def to_dec_coin(self) -> Coin:
        return Coin(self.denom, Dec(self.amount))

    def __str__(self) -> str:
        if self.is_dec_coin():
            amount_str = str(self.amount).rstrip("0")
            if amount_str.endswith("."):
                amount_str += "0"
            return f"{amount_str}{self.denom}"
        return f"{self.amount}{self.denom}"

    def to_data(self) -> dict:
        return {"denom": self.denom, "amount": str(self.amount)}

    @classmethod
    def from_str(cls, string: str) -> Coin:
        pattern = r"^(\-?[0-9]+(\.[0-9]+)?)([a-zA-Z]+)$"
        match = re.match(pattern, string)
        if match is None:
            raise ValueError(f"failed to parse Coin: {string}")
        else:
            return cls(match.group(3), match.group(1))

    def add(self, other: Union[Numeric.Input, Coin]) -> Coin:
        if isinstance(other, Coin):
            if other.denom != self.denom:
                raise ArithmeticError(
                    f"cannot add/subtract two Coin objects of different denoms: {self.denom} and {other.denom}"
                )
            return Coin(self.denom, self.amount + other.amount)
        else:
            return Coin(self.denom, self.amount + Numeric.parse(other))

    def __add__(self, other: Union[Numeric.Input, Coin]) -> Coin:
        return self.add(other)

    def sub(self, other: Union[Numeric.Input, Coin]) -> Coin:
        if isinstance(other, Coin):
            return self.add(other.mul(-1))
        else:
            return self.add(Numeric.parse(other) * -1)

    def __sub__(self, other: Union[Numeric.Input, Coin]) -> Coin:
        return self.sub(other)

    def mul(self, other: Numeric.Input) -> Coin:
        other_amount = Numeric.parse(other)
        return Coin(self.denom, self.amount * other_amount)

    def __mul__(self, other: Numeric.Input) -> Coin:
        return self.mul(other)

    def div(self, other: Numeric.Input) -> Coin:
        other_amount = Numeric.parse(other)
        if isinstance(other_amount, int):
            return Coin(self.denom, (self.amount // other_amount))
        else:
            return Coin(self.denom, (self.amount / other_amount))

    def __truediv__(self, other: Numeric.Input) -> Coin:
        return self.div(other)

    def __floordiv__(self, other: Numeric.Input) -> Coin:
        return self.div(int(Numeric.parse(other)))

    def mod(self, other: Numeric.Input) -> Coin:
        other_amount = Numeric.parse(other)
        if isinstance(other_amount, Dec):
            return Coin(self.denom, Dec(self.amount).mod(other_amount))
        else:
            return Coin(self.denom, self.amount % other_amount)

    def __mod__(self, other: Numeric.Input) -> Coin:
        return self.mod(other)

    def __neg__(self) -> Coin:
        return Coin(denom=self.denom, amount=(-self.amount))

    def __abs__(self) -> Coin:
        return Coin(denom=self.denom, amount=abs(self.amount))

    def __pos__(self) -> Coin:
        return abs(self)

    @classmethod
    def from_data(cls, data: dict) -> Coin:
        return cls(data["denom"], data["amount"])



from typing import List

from terra_sdk.core.gov import Proposal

from ._base import BaseAPI


class AsyncGovAPI(BaseAPI):
    async def proposals(self) -> List[Proposal]:
        res = await self._c._get("/gov/proposals")
        return [Proposal.from_data(d) for d in res]

    async def proposal(self, proposal_id: int) -> Proposal:
        res = await self._c._get("/gov/proposals/{proposal_id}")
        return Proposal.from_data(res)

    async def proposer(self, proposal_id: int) -> str:
        res = await self._c._get(f"/gov/proposals/{proposal_id}/proposer")
        return res["proposer"]

    async def deposits(self, proposal_id: int):
        return await self._c._get(f"/gov/proposals/{proposal_id}/deposits")

    async def votes(self, proposal_id: int):
        return await self._c._get(f"/gov/proposals/{proposal_id}/votes")

    async def tally(self, proposal_id: int):
        return await self._c._get(f"/gov/proposals/{proposal_id}/tally")

    async def deposit_parameters(self) -> dict:
        return await self._c._get("/gov/parameters/deposit")

    async def voting_parameters(self) -> dict:
        return await self._c._get("/gov/parameters/voting")

    async def tally_parameters(self) -> dict:
        return await self._c._get("/gov/parameters/tallying")

    async def parameters(self) -> dict:
        return {
            "deposit_params": await self.deposit_parameters(),
            "voting_params": await self.voting_parameters(),
            "tally_params": await self.tally_parameters(),
        }


class GovAPI(BaseAPI):
    def proposals(self) -> List[Proposal]:
        """Fetches all proposals.

        Returns:
            List[Proposal]: proposals
        """
        res = self._c._get("/gov/proposals")
        return [Proposal.from_data(d) for d in res]

    def proposal(self, proposal_id: int) -> Proposal:
        """Fetches a single proposal by id.

        Args:
            proposal_id (int): proposal ID

        Returns:
            Proposal: proposal
        """
        res = self._c._get("/gov/proposals/{proposal_id}")
        return Proposal.from_data(res)

    def proposer(self, proposal_id: int) -> str:
        """Fetches the proposer of a proposal.

        Args:
            proposal_id (int): proposal ID

        Returns:
            str: proposal's proposer
        """
        res = self._c._get(f"/gov/proposals/{proposal_id}/proposer")
        return res["proposer"]

    def deposits(self, proposal_id: int):
        """Fetches the deposit information about a proposal.

        Args:
            proposal_id (int): proposal ID
        """
        return self._c._get(f"/gov/proposals/{proposal_id}/deposits")

    def votes(self, proposal_id: int):
        """Fetches the votes for a proposal.

        Args:
            proposal_id (int): proposal ID
        """
        return self._c._get(f"/gov/proposals/{proposal_id}/votes")

    def tally(self, proposal_id: int):
        """Fetches the tally for a proposal.

        Args:
            proposal_id (int): proposal ID
        """
        return self._c._get(f"/gov/proposals/{proposal_id}/tally")

    def deposit_parameters(self) -> dict:
        """Fetches the Gov module's deposit parameters.

        Returns:
            dict: deposit parameters
        """
        return self._c._get("/gov/parameters/deposit")

    def voting_parameters(self) -> dict:
        """Fetches the Gov module's voting parameters.

        Returns:
            dict: voting parameters
        """
        return self._c._get("/gov/parameters/voting")

    def tally_parameters(self) -> dict:
        """Fetches the Gov module's tally parameters.

        Returns:
            dict: tally parameters
        """
        return self._c._get("/gov/parameters/tallying")

    def parameters(self) -> dict:
        """Fetches the Gov module's parameters.

        Returns:
            dict: Gov module parameters
        """
        return {
            "deposit_params": self.deposit_parameters(),
            "voting_params": self.voting_parameters(),
            "tally_params": self.tally_parameters(),
        }
